import { EventEmitter } from 'events';
import createRequestClient from '../create-request-client';
import { transform, transformBatch } from '../mapping-kit';
import { fieldsToJsonSchema } from './fields-to-jsonschema';
import { Response } from '../fetch';
import { syncModeTypes } from './types';
import { validateSchema } from '../schema-validation';
import { IntegrationError } from '../errors';
import { removeEmptyValues } from '../remove-empty-values';
export const hookTypeStrings = ['onMappingSave', 'retlOnMappingSave'];
const isSyncMode = (value) => {
    return syncModeTypes.find((validValue) => value === validValue) !== undefined;
};
export class Action extends EventEmitter {
    constructor(destinationName, definition, extendRequest) {
        super();
        this.definition = definition;
        this.destinationName = destinationName;
        this.extendRequest = extendRequest;
        this.hasBatchSupport = typeof definition.performBatch === 'function';
        this.hasHookSupport = definition.hooks !== undefined;
        if (Object.keys(definition.fields ?? {}).length) {
            this.schema = fieldsToJsonSchema(definition.fields);
        }
        if (definition.hooks) {
            for (const hookName in definition.hooks) {
                const hook = definition.hooks[hookName];
                if (hook?.inputFields) {
                    if (!this.hookSchemas) {
                        this.hookSchemas = {};
                    }
                    const castedInputFields = {};
                    for (const key in hook.inputFields) {
                        const field = hook.inputFields[key];
                        if (field.dynamic) {
                            castedInputFields[key] = {
                                ...field,
                                dynamic: true
                            };
                        }
                        else {
                            castedInputFields[key] = {
                                ...field,
                                dynamic: false
                            };
                        }
                    }
                    this.hookSchemas[hookName] = fieldsToJsonSchema(castedInputFields);
                }
            }
        }
    }
    async execute(bundle) {
        const results = [];
        let payload = transform(bundle.mapping, bundle.data);
        results.push({ output: 'Mappings resolved' });
        payload = removeEmptyValues(payload, this.schema, true);
        if (bundle.mapping && '__segment_internal_sync_mode' in bundle.mapping) {
            delete payload['__segment_internal_sync_mode'];
        }
        if (this.schema) {
            const schemaKey = `${this.destinationName}:${this.definition.title}`;
            validateSchema(payload, this.schema, { schemaKey, statsContext: bundle.statsContext });
            results.push({ output: 'Payload validated' });
        }
        let hookOutputs = {};
        if (this.definition.hooks) {
            for (const hookType in this.definition.hooks) {
                const hookOutputValues = bundle.mapping?.[hookType];
                if (hookOutputValues) {
                    hookOutputs = { ...hookOutputs, [hookType]: hookOutputValues };
                }
            }
        }
        const syncMode = this.definition.syncMode ? bundle.mapping?.['__segment_internal_sync_mode'] : undefined;
        const dataBundle = {
            rawData: bundle.data,
            rawMapping: bundle.mapping,
            settings: bundle.settings,
            payload,
            auth: bundle.auth,
            features: bundle.features,
            statsContext: bundle.statsContext,
            logger: bundle.logger,
            dataFeedCache: bundle.dataFeedCache,
            transactionContext: bundle.transactionContext,
            stateContext: bundle.stateContext,
            audienceSettings: bundle.audienceSettings,
            hookOutputs,
            syncMode: isSyncMode(syncMode) ? syncMode : undefined
        };
        const output = await this.performRequest(this.definition.perform, dataBundle);
        results.push({ data: output, output: 'Action Executed' });
        return results;
    }
    async executeBatch(bundle) {
        const results = [{ output: 'Action Executed' }];
        if (!this.hasBatchSupport) {
            throw new IntegrationError('This action does not support batched requests.', 'NotImplemented', 501);
        }
        let payloads = transformBatch(bundle.mapping, bundle.data);
        if (bundle.mapping && '__segment_internal_sync_mode' in bundle.mapping) {
            for (const payload of payloads) {
                if (payload) {
                    delete payload['__segment_internal_sync_mode'];
                }
            }
        }
        if (this.schema) {
            const schema = this.schema;
            const validationOptions = {
                schemaKey: `${this.destinationName}:${this.definition.title}`,
                throwIfInvalid: false,
                statsContext: bundle.statsContext
            };
            payloads = payloads
                .map((payload) => removeEmptyValues(payload, schema))
                .filter((payload) => validateSchema(payload, schema, validationOptions));
        }
        let hookOutputs = {};
        if (this.definition.hooks) {
            for (const hookType in this.definition.hooks) {
                const hookOutputValues = bundle.mapping?.[hookType];
                if (hookOutputValues) {
                    hookOutputs = { ...hookOutputs, [hookType]: hookOutputValues };
                }
            }
        }
        if (payloads.length === 0) {
            return results;
        }
        if (this.definition.performBatch) {
            const syncMode = this.definition.syncMode ? bundle.mapping?.['__segment_internal_sync_mode'] : undefined;
            const data = {
                rawData: bundle.data,
                rawMapping: bundle.mapping,
                settings: bundle.settings,
                audienceSettings: bundle.audienceSettings,
                payload: payloads,
                auth: bundle.auth,
                features: bundle.features,
                statsContext: bundle.statsContext,
                logger: bundle.logger,
                dataFeedCache: bundle.dataFeedCache,
                transactionContext: bundle.transactionContext,
                stateContext: bundle.stateContext,
                hookOutputs,
                syncMode: isSyncMode(syncMode) ? syncMode : undefined
            };
            const output = await this.performRequest(this.definition.performBatch, data);
            results[0].data = output;
            return results;
        }
        return results;
    }
    async executeDynamicField(field, data, dynamicFn) {
        let fn;
        if (dynamicFn && typeof dynamicFn === 'function') {
            fn = dynamicFn;
        }
        else {
            fn = this.definition.dynamicFields?.[field];
        }
        if (typeof fn !== 'function') {
            return Promise.resolve({
                choices: [],
                nextPage: '',
                error: {
                    message: `No dynamic field named ${field} found.`,
                    code: '404'
                }
            });
        }
        return (await this.performRequest(fn, data));
    }
    async executeHook(hookType, data) {
        if (!this.hasHookSupport) {
            throw new IntegrationError('This action does not support any hooks.', 'NotImplemented', 501);
        }
        const hookFn = this.definition.hooks?.[hookType]?.performHook;
        if (!hookFn) {
            throw new IntegrationError(`Missing implementation for hook: ${hookType}.`, 'NotImplemented', 501);
        }
        if (this.hookSchemas?.[hookType]) {
            const schema = this.hookSchemas[hookType];
            validateSchema(data.hookInputs, schema);
        }
        return (await this.performRequest(hookFn, data));
    }
    async performRequest(requestFn, data) {
        const requestClient = this.createRequestClient(data);
        const response = await requestFn(requestClient, data);
        return this.parseResponse(response);
    }
    createRequestClient(data) {
        const options = this.extendRequest?.(data) ?? {};
        return createRequestClient(options, {
            afterResponse: [this.afterResponse.bind(this)],
            statsContext: data.statsContext
        });
    }
    afterResponse(request, options, response) {
        const modifiedResponse = response;
        modifiedResponse.request = request;
        modifiedResponse.options = options;
        this.emit('response', modifiedResponse);
        return modifiedResponse;
    }
    parseResponse(response) {
        if (response instanceof Response) {
            return response.data ?? response.content;
        }
        return response;
    }
}
//# sourceMappingURL=action.js.map